#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "graph.h"
#include "fastme.h"

/*functions from bme.c*/
void makeBMEAveragesTable(tree *T, double **D, double **A);
void assignBMEWeights(tree *T, double **A);

/*from gme.c*/
void makeOLSAveragesTable(tree *T, double **D, double **A);

/*from graph.c*/
edge *siblingEdge(edge *e);
void weighTree(tree *T);

/*from traverse.c*/
edge *depthFirstTraverse(tree *T, edge *e);

edge *threeWayMin(edge *e1, edge *e2, edge *e3, double *v)
{
  if (v[e1->head->index] < v[e2->head->index])
    if (v[e1->head->index] < v[e3->head->index])
      return(e1);
  if (v[e2->head->index] < v[e3->head->index])
    return(e2);
  return(e3);
}


double pathAvg(node *v, edge *afixed, edge *pathup, node *lca, double **A)
     /*average from node v to all trees off path from pathup to lca and 
       also off path from afixed to lca*/
     /*afixed is far end of path, pathup is near end*/
{
  double c, sum;
  edge *eloop, *floop;
  sum = 0.0;
  if (NULL != pathup)
    {
      c = 0.5;
      eloop = pathup;
      while (eloop->tail != lca) /*going up path ck-1,...c1*/
	{
	  floop = siblingEdge(eloop);
	  sum += c*A[v->index][floop->head->index];
	  eloop = eloop->tail->parentEdge;
	  c = 0.5*c;
	}
      /*case where afixed is at top of path up, either above
	path or skew to path*/
      if ((afixed->head == lca) || (afixed->tail == lca))
	{
	  c = 2.0*c;
	  sum += c*A[v->index][afixed->head->index];
	  return(sum);
	}
      /*else*/
      sum += c*A[v->index][lca->index];
    }
  else /*NULL == pathup*/
    c = 1.0;
  eloop = afixed->tail->parentEdge;
  while (eloop->tail != lca)
    {
      c = 0.5*c;
      eloop = eloop->tail->parentEdge;
    }
  sum += c*A[v->index][afixed->head->index];
  eloop = afixed->tail->parentEdge;
  while (eloop->tail != lca)
    {
      floop = siblingEdge(eloop);
      sum += c*A[v->index][floop->head->index];
      /*note that the smallest value of c is used twice, 
	other values used exactly once*/
      c = 2.0*c;
      eloop = eloop->tail->parentEdge;
    }
  return(sum);
}

void pruneSubtree(edge *p, edge *u, edge *d)
     /*removes p, d from tree, u connects to d->head to compensate*/
{
  p->tail->parentEdge = NULL;/*remove p subtree*/
  u->head = d->head;
  d->head->parentEdge = u;  /*u connected to d->head*/
  d->head = NULL; /*d removed from tree*/  
}

void SPRsplitEdge(edge *e, edge *p, edge *d)
     /*splits edge e to make it parental to p,d.  d is parental to what
       previously was below e*/
{
  d->head = e->head;
  e->head = p->tail;
  p->tail->parentEdge = e;
  d->head->parentEdge = d;
}

void SPRrotate(edge *eout, node *stopnode, int leftOrRight, edge **elist)
{
  edge *eup, *emove;
  node *oldPivot, *newPivot;
  oldPivot = eout->head;
  if (LEFT == leftOrRight) /*pushing tree down to right*/
    /*split edge is in left subtree*/
      eup = oldPivot->leftEdge;
  else/*RIGHT*/
      eup = oldPivot->rightEdge;
  if (stopnode == eup->head)
    return;
  emove = siblingEdge(elist[0]);
  newPivot = eup->head;
  newPivot->parentEdge = eout;
  oldPivot->parentEdge = eup;
  eup->head = oldPivot;
  eout->head = newPivot;
  eup->tail = newPivot;
  if (LEFT == leftOrRight)
    oldPivot->leftEdge = emove;
  else
    oldPivot->rightEdge = emove;
  emove->tail = oldPivot;
  if (elist[0] == newPivot->leftEdge)
    {
      newPivot->rightEdge = eup;
      SPRrotate(eout,stopnode,LEFT,elist+1);
    }
  else
    {
      newPivot->leftEdge = eup;
      SPRrotate(eout,stopnode,RIGHT,elist+1);
    }
}
    
	  


/*topological shift function*/
void SPRshift(tree *T, node *v, edge *e, double **D, double **A)
{
  edge *vup, *vdown, *vpar;
  vpar = v->parentEdge;
  vdown = siblingEdge(vpar);
  vup = vpar->tail->parentEdge;
  /*topological shift*/
  pruneSubtree(vpar,vup,vdown); /*removes v subtree and vdown, extends vup*/
  SPRsplitEdge(e,vpar,vdown); /*splits e to make e sibling edge to vpar,
				both below vup*/
  /*recalculate A*/
  makeBMEAveragesTable(T,D,A);
}
      
edge *optimalBalSPRDown(tree *T, node *v, edge *topEdge, double **A, double *val);

edge *optimalBalSPRUp(tree *T, node *v, double **A, double *val);

edge *optimalBalSPRBelow(tree *T, node *v, double **A, double *val, int *lr);

edge *optimalBalSPRUpTreeShift(tree *T, node *v, double **A, double *val, edge *etest);



/*the SPR function tests all the possible SPR topology rearrangements.  In an unrooted tree, a SPR
consists of removing a subtree S, contracting the two edges that met where the S met T, and splitting
a new edge, adding S in the new location.  In a rooted tree, there are two kinds of subtrees: those above an internal
and those below an internal edge*/
void SPR(tree *T, double **D, double **A, int *count)
{
  double SPRvalue = 0.0;
  double SPRnow;
  double oldweight = 0.0;
  int UpOrDown=0;
  int LeftOrRight=0;
  int LRmove;
  int pathLength;
  boolean noSwap = TRUE;
  edge *e, *f, *sib, *esplit;
  edge **elist;
  node *vtest, *vmove;
  do
    { /*esplit loop*/
      esplit = NULL;
      vmove = NULL;
      for(e=depthFirstTraverse(T,NULL);T->root!=e->tail;e=depthFirstTraverse(T,e))
	{
	  /*in this loop, we test the possibility of moving the subtree below e->head*/
	  vtest = e->head;
	  sib = siblingEdge(e);
	  f = optimalBalSPRDown(T,vtest,sib,A,&SPRnow);	  
	  /*the subtree can be moved into a position below the sibling edge, or
	    into a position in the tree above the parental edge*/
	  /*first possibility is to move vtest subtree into tree below edge called sib*/
	  /*f is the optimal edge in the subtree below sib*/
	  if (SPRnow < SPRvalue)
	    {
	      UpOrDown=DOWN;
	      vmove = vtest;
	      esplit = f;
	      SPRvalue = SPRnow;
	    }
	  
	  f = optimalBalSPRUp(T,vtest,A,&SPRnow);
	  /*second possibility is to move vtest into tree above vtest, sib subtree*/
	  /*f is the optimal edge in the subtree above vtest*/
	  
	  if (SPRnow < SPRvalue)
	    {
	      UpOrDown=DOWN;
	      vmove = vtest;
	      esplit = f;
	      SPRvalue = SPRnow;
	      }
	  
	  f = optimalBalSPRBelow(T, vtest, A, &SPRnow,&LeftOrRight);

	  /*The other possibility is that the subtree is above vtest, not below it*/
	  /* in this scenario, f is the optimal edge below vtest*/
	  if (SPRnow < SPRvalue)
	    {
	      UpOrDown = UP;
	      vmove = vtest;
	      esplit = f;
	      LRmove = LeftOrRight;
	      SPRvalue = SPRnow;
	    }
	} /*e search*/
      if (NULL != esplit)
	{
	  assignBMEWeights(T,A);
	  weighTree(T);
	  if (verbose)
	    {
	      printf("SPR:Shifting %s to split %s and %s\n",vmove->label,
		     esplit->tail->label,esplit->head->label);
	      printf("SPR:Old tree length is %lf.\n",T->weight);
	      printf("SPR:New tree length should be %lf.\n",T->weight + 0.25*SPRvalue);
	      if (DOWN == UpOrDown)
		printf("SPR:Shifting tree below %s\n",vmove->label);
	      else
		printf("SPR:Shifting tree above %s\n",vmove->label);
	    }
	  oldweight = T->weight;
	  if (DOWN == UpOrDown)
	    SPRshift(T,vmove,esplit,D,A);
	  else /*shifting a subtree containing the root*/
	    {
	      for(f=esplit,pathLength=0;f->tail != vmove;f=f->tail->parentEdge)
		pathLength++;
	      elist = (edge **)malloc(pathLength*sizeof(edge *));
	      /*note that pathLength > 0 will hold*/
	      f = esplit;
	      while (pathLength > 0)		
		{
		  pathLength--;
		  elist[pathLength] = f;
		  f = f->tail->parentEdge;
		}
	      SPRrotate(vmove->parentEdge,esplit->head,LRmove,elist);
	      makeBMEAveragesTable(T,D,A);
	      free(elist);
	    }
	  assignBMEWeights(T,A);
	  weighTree(T);
	  if (verbose)
	    printf("SPR:new tree length is %lf.\n",T->weight);
	  if (T->weight > oldweight)
	    {
	      fprintf(stderr,"Error in SPR routine!\n");
	      exit(EXIT_FAILURE);
	    }
	  (*count)++;

	}
      SPRvalue = 0.0;
    } while (NULL != esplit);
}
  
edge *optimalSPRBalDownsubtree(edge *esplit, tree *T, node *vtest, edge *top,
			       double **A, double *swapValues);

/*seeks optimal edge in subtree with top edge topEdge, where value for
  each edge is given by iteratation of NNI tree swap formula */

edge *optimalBalSPRDown(tree *T, node *vtest, edge *topEdge, double **A, double *val)
     /*
       this function returns the edge e below topEdge that is the optimal 
       insertion point for the node vtest*/
     /*topEdge is sibling edge to parent edge of vtest*/
{
  double *swapValues;
  edge *left, *right;
  left = topEdge->head->leftEdge;
  right = topEdge->head->rightEdge;
  if (NULL == left)  /*no subtree below topEdge.  Splitting topEdge returns
		       value of 0.0*/
    {
      *val = 0.0;
      return(NULL);
    }
  else /*search subtree*/
    {
      /*splitting topEdge is equivalent to performing no SPR*/
      swapValues = (double *)malloc(T->size*sizeof(double));
      swapValues[topEdge->head->index] = 0.0;
      left = optimalSPRBalDownsubtree(left,T,vtest,topEdge,A,swapValues);
      right = optimalSPRBalDownsubtree(right,T,vtest,topEdge,A,swapValues);
      if (swapValues[left->head->index] < swapValues[right->head->index])
	{
	  if (swapValues[left->head->index] < -1.0*EPSILON)
	    {
	      *val = swapValues[left->head->index];
	      free(swapValues);
	      return(left);
	    }
	  else /*no good SPRs in subtree*/
	    {
	      *val = 0.0;
	      free(swapValues);
	      return(NULL);
	    }
	}
      else if (swapValues[right->head->index] < -1.0*EPSILON)
	{
	  *val = swapValues[right->head->index];
	  free(swapValues);
	  return(right);
	}
      else /*no good SPRs in subtree*/
	{
	  *val = 0.0;
	  free(swapValues);
	  return(NULL);
	}
    }
}



edge *optimalSPRBalDownsubtree(edge *etest, tree *T, node *vtest, edge *top,
			   double **A, double *swapValues)
     /*returns edge optimal in tree below etest, when finding insertion point 
       for vtest (and tree below vtest).  the edge top is the sibling edge
       for vtest's parent, and it is above etest*/
{
  edge *left, *right;
  edge *ck;

  ck = siblingEdge(etest);
  
  /*formula: Given swapping subtree B to split etest, w/bottom subtree D, subtree
    A above topEdge, and C_i subtrees from A to e,
    difference from value of swap e's parent edge to a swap at e is 
    D_{A*|C_k} + D_{B|D} - D_{A*|B} - D_{C_k|D},
    where A* is union of A with all C_i with i < k, and C_k is subtree
    nearest to etest*/
  
  swapValues[etest->head->index] = swapValues[etest->tail->index]
    + A[vtest->index][etest->head->index] 
    - A[ck->head->index][etest->head->index];
  /*simplest terms first: D_{B|D} and -D_{C_k|D}*/

  /*now, add in D_{A*|C_k} and -D_{A*|B}*/
  

  swapValues[etest->head->index] += pathAvg(ck->head,top->tail->parentEdge,etest->tail->parentEdge,top->tail,A);
  swapValues[etest->head->index] -= pathAvg(vtest,top->tail->parentEdge,etest->tail->parentEdge,top->tail,A);
  left = etest->head->leftEdge;
  if (NULL != left) /*recursive part*/
    {
      right = etest->head->rightEdge;
      left = optimalSPRBalDownsubtree(left,T,vtest,top,A,swapValues);
      right = optimalSPRBalDownsubtree(right,T,vtest,top,A,swapValues);
      return(threeWayMin(left,right,etest,swapValues));
    }
  else /*left is NULL*/
    return(etest);
}

edge *optimalBalSPRSkewsubtree(edge *etest, tree *T, node *vtest,edge *afixed,
			      double **A, double *swapValues);
edge *optimalBalSPRUpsubtree(edge *etest, tree *T, node *vtest,
			     double **A, double *swapValues, edge *eback,
			     edge *afixed, edge *par);

edge *optimalBalSPRUp(tree *T, node *vtest, double **A, double *val)
     /*vtest is node being moved, search tree above vtest->parentEdge->tail->parentEdge*/
{
  double *swapValues = NULL;
  edge *up, *skew, *par, *afixed;
  /*in this scheme, fixed tree is below out, B is below sibling edge of
    out*/
  afixed = siblingEdge(vtest->parentEdge);
  par = afixed->tail->parentEdge;
  up = par->tail->parentEdge;
  if (NULL == up) /*no tree above par, no swaps to test*/
    {
      *val = 0.0;
      return(NULL);
    }
  else 
    { /*search up*/
      skew = siblingEdge(par);
      swapValues = (double *)malloc(T->size*sizeof(double));
      swapValues[par->head->index] = 0.0;
      up = optimalBalSPRUpsubtree(up,T,vtest,A,swapValues,par,afixed,par);
      /*optimal edge in tree */
      skew = optimalBalSPRSkewsubtree(skew,T,vtest,afixed,A,swapValues);
      if (swapValues[up->head->index] < swapValues[skew->head->index])
	{
	  if (swapValues[up->head->index] < -1.0*EPSILON)
	    {
	      *val = swapValues[up->head->index];
	      free(swapValues);
	      return(up);
	    }
	  else /*no swaps that decrease tree length*/
	    {
	      *val = 0.0;
	      free(swapValues);
	      return(NULL);
	    }
	}
      else if (swapValues[skew->head->index] < -1.0*EPSILON)
	{
	  *val = swapValues[skew->head->index];
	  free(swapValues);
	  return(skew);
	}
      else /*up, skew sub-optimal*/
	{
	  free(swapValues);
	  *val = 0.0;
	  return(NULL);
	}
    }
}

edge *optimalBalSPRSkewsubsubtree(node *lca, edge *etest, tree *T, node *vtest,edge *afixed,
				  double **A, double *swapValues);

/*top level only of skew subtree*/
edge *optimalBalSPRSkewsubtree(edge *etest, tree *T, node *vtest, edge *afixed,
			       double **A, double *swapValues)
     
     /*etest is edge to be split, its tail is above vtest, but the 
       edge itself isn't*/
     
{
  edge *left, *right;
  edge *eback;
  eback = siblingEdge(etest); /*sib is on path to vtest's original location*/

  /*formula: Given swapping B to split etest, with bottom subtree D, subtree
    A below Afixed, and C_i subtrees from A to etest,
    difference from value of swap at eback to a swap at etest is
    D_{A*|C_k} + D_{B|D} - D_{A*|B} - D_{C_k|D},
    where A* is union of A with all C_i with i < k, and C_k is subtree
    above etest->tail*/

  /*simplest terms first: D_{B|D} and -D_{C_k|D}*/  
  swapValues[etest->head->index] = swapValues[eback->head->index]
    + A[vtest->index][etest->head->index] - A[etest->head->index][etest->tail->index];

  /*now, add in D_{A*|C_k} and -D_{A*|B}*/
  
  swapValues[etest->head->index] += pathAvg(etest->tail,afixed,NULL,etest->tail,A);
  swapValues[etest->head->index] -= pathAvg(vtest,afixed,NULL,etest->tail,A);
  

  left = etest->head->leftEdge;
  if (NULL != left)
    {
      right = etest->head->rightEdge;
      left = optimalBalSPRSkewsubsubtree(etest->tail,left,T,vtest,afixed,A,swapValues);
      right = optimalBalSPRSkewsubsubtree(etest->tail,right,T,vtest,afixed,A,swapValues);
      return(threeWayMin(left,right,etest,swapValues));
    }
  else /*left is NULL*/
    return(etest);
}


/*test spliting e when path from vtest to e goes up from e->tail to 
  near before going down to out*/
edge *optimalBalSPRSkewsubsubtree(node *lca, edge *etest, tree *T, node *vtest, 
				  edge *afixed, double **A, double *swapValues)
{
  edge *left, *right;
  edge *ck;
  ck = siblingEdge(etest);
  /*formula: Given swapping B to split etest, with bottom subtree D, subtree
    A below afixed (next to B originally), and C_i subtrees from A to etest,
    difference from value of swap e's parent edge to a swap at e is
    D_{A*|C_k} + D_{B|D} - D_{A*|B} - D_{C_k|D},
    where A* is union of A with all C_i with i < k, and C_k is subtree
    nearest to e*/
  
  /*simplest terms first: D_{B|D} and -D_{C_k|D}*/
  swapValues[etest->head->index] = swapValues[etest->tail->index]
    + A[vtest->index][etest->head->index] - A[ck->head->index][etest->head->index];

  /*now, add in D_{A*|C_k} and -D_{A*|B}*/
  swapValues[etest->head->index] += pathAvg(ck->head,afixed,etest->tail->parentEdge,lca,A);
  swapValues[etest->head->index] -= pathAvg(vtest,afixed,etest->tail->parentEdge,lca,A);

  left = etest->head->leftEdge;
  if (NULL != left)
    {
      right = etest->head->rightEdge;
      left = optimalBalSPRSkewsubsubtree(lca,left,T,vtest,afixed,A,swapValues);
      right = optimalBalSPRSkewsubsubtree(lca,right,T,vtest,afixed,A,swapValues);
      return(threeWayMin(left,right,etest,swapValues));
    }
  else /*left is NULL*/
    return(etest);
}

edge *optimalBalSPRUpsubtree(edge *etest, tree *T, node *vtest,
			     double **A, double *swapValues, edge *eback,
			     edge *afixed, edge *par)
{
  /*testing split of etest*/
  /*eback is first edge on path leading back down from etest
    to the old lcation of vtest*/
  edge *ck, *up, *skew;

  ck = siblingEdge(eback);


  /*formula: Given swapping B to split etest, A below sib, 
    D above topEdge, and C_i subtrees from A to e,
    difference from value of swap e's parent edge to a swap at e is
    D_{A*|C_k} + D_{B|D} - D_{A*|B} - D_{C_k|D},
    where A* is union of A with all C_i with i < k, and C_k is subtree
    nearest to e*/    

  /*In this case, D is tree above etest, C_k is sibling tree to near,
    B is the SPR subtree (as always), A is tree below afixed and 
    C_i are trees along path from afixed to near*/

  /*simplest terms first: D_{B|D} and -D_{C_k|D}*/
  swapValues[etest->head->index] = swapValues[eback->head->index]
    + A[vtest->index][etest->head->index] - A[ck->head->index][etest->head->index];


  /*now, add in D_{A*|C_k} and -D_{A*|B}*/
  
  swapValues[etest->head->index] += pathAvg(ck->head,afixed,NULL,eback->tail,A);
  swapValues[etest->head->index] -= pathAvg(vtest,afixed,NULL,eback->tail,A);


  up = etest->tail->parentEdge;
  if (NULL != up)
    {
      skew = siblingEdge(etest);
      up = optimalBalSPRUpsubtree(up,T,vtest,A,swapValues,etest,afixed,par);
      skew = optimalBalSPRSkewsubtree(skew,T,vtest,afixed,A,swapValues);
      return(threeWayMin(up,skew,etest,swapValues));
    }
  else /*up is NULL*/
    return(etest);
}

edge *optimalBalSPRBelowsubtree(edge *etest, tree *T, node *vtest,double **A,
				double *swapValues, edge *afixed);

edge *optimalBalSPRBelow(tree *T, node *v, double **A, double *val, int *lr)
{
  double *swapValues;
  edge *left, *right, *lfixed, *rfixed;
  rfixed = left = v->leftEdge;
  lfixed = right = v->rightEdge;
  if (NULL == left)
    {
      *val = 0.0;
      return(NULL);
    }
  swapValues = (double *)malloc(T->size*sizeof(double));
  swapValues[left->head->index] = swapValues[right->head->index] = 0.0;
  left = optimalBalSPRBelowsubtree(left,T,v,A,swapValues,lfixed);
  right = optimalBalSPRBelowsubtree(right,T,v,A,swapValues,rfixed);
  if (swapValues[left->head->index] < swapValues[right->head->index])
    if (swapValues[left->head->index] < -1.0*EPSILON)
      {
	*val = swapValues[left->head->index];
	free(swapValues);
	*lr = LEFT;
	return(left);
      }
  else
    {
      *val = 0.0;
      free(swapValues);
      return(NULL);
    }
  else if (swapValues[right->head->index] < -1.0*EPSILON)
    {
      *val = swapValues[right->head->index];
      free(swapValues);
      *lr = RIGHT;
      return(right);
    }
  else
    {
      *val = 0.0;
      free(swapValues);
      return(NULL);
    }
}

edge *optimalBalSPRBelowsubtree(edge *etest, tree *T, node *vtest,double **A,
				double *swapValues, edge *afixed)
{
  edge *ck, *left, *right;

  if (vtest == etest->tail)
    swapValues[etest->head->index] = 0.0;
  else  
    {
      ck = siblingEdge(etest);  
      /*formula: Given swapping B to split etest, A below afixed, 
	D below etest, and C_i subtrees from A to etest,
	difference from value of swap at etest's parent edge to a swap at etest 
	is D_{A*|C_k} + D_{B|D} - D_{A*|B} - D_{C_k|D},
	where A* is union of A with all C_i with i < k, and C_k is subtree
	nearest to e*/    
      
      
      /*simplest terms first: D_{B|D} and -D_{C_k|D}*/
      swapValues[etest->head->index] = swapValues[etest->tail->index]
	+ A[vtest->index][etest->head->index] - A[ck->head->index][etest->head->index];
      
      /*now, add in D_{A*|C_k} and -D_{A*|B}*/
      
      swapValues[etest->head->index] += pathAvg(ck->head,afixed,etest->tail->parentEdge,vtest,A);
      swapValues[etest->head->index] -= pathAvg(vtest,afixed,etest->tail->parentEdge,vtest,A);
    }
  left = etest->head->leftEdge;
  right = etest->head->rightEdge;
  if (NULL == left)
    return(etest);
  else
    {
      left = optimalBalSPRBelowsubtree(left,T,vtest,A,swapValues,afixed);
      right = optimalBalSPRBelowsubtree(right,T,vtest,A,swapValues,afixed);
      if (swapValues[left->head->index] < swapValues[right->head->index])
	{
	  if (swapValues[left->head->index] < swapValues[etest->head->index])
	    return(left);
	  else
	    return(etest);
	}
      else if (swapValues[right->head->index] < swapValues[etest->head->index])
	return(right);
      else
	return(etest);
    }
  
}

