#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "fastme.h"
#include "graph.h"
#include "newick.h"
#include <string.h>
#include "distance.h"

/*functions from newick.c*/
tree *loadNewickTree(FILE *ifile, int numLeaves);
void NewickPrintTree(tree *T, FILE *ofile);
tree *detrifurcate(tree *T);
void partitionSizes(tree *T);

/*functions from bme.c*/
tree *BMEaddSpecies(tree *T,node *v, double **D, double **A); 
void makeBMEAveragesTable(tree *T, double **D, double **A);
void assignBMEWeights(tree *T, double **A);

/*from gme.c*/
tree *GMEaddSpecies(tree *T,node *v, double **D, double **A);
void makeOLSAveragesTable(tree *T, double **D, double **A);
void assignOLSWeights(tree *T, double **A);
void assignAllSizeFields(tree *T);

/*from nj.c*/
tree *nj(double **D, set *species, int n);
int *initZeroArray(int l);
int *initOneArray(int l);

/*from inputs.c*/
double **loadMatrix(FILE *ifile, int *size, set *S);
char **loadSequences(FILE *ifile, int *size, set  *S, int *seqLength);
void compareSets(tree *T, set *S, FILE *ofile);
void freeMatrix(double **D, int size);
double **loadScoreMatrix(int *size, FILE *sfile, char *alphabet);
void freeCharMatrix(char **M, int numStrings);

/*from print.c*/
void printMatrix(double **D, int size, set *nodes, FILE *ofile);

/*from NNI.c*/
void NNI(tree *T, double **avgDistArray, int *count);

/*from bNNI.c*/
void bNNI(tree *T, double **avgDistArray, int *count);

/*from SPR.c*/
void SPR(tree *T, double **D, double **A, int *count);

/*from graph.c*/
void freeSet(set *S);
void freeTree(tree *T);
node *copyNode(node *v);

/*from distance.c*/
double **makeDistMatrix(char **data, int numSeqs, int numSites, int model, int *filter, double gamma, double **scoreMatrix,
			int alphaSize, char *alphabet);
void gapCheckFilter(int *filter, int itype, int seqlength, int numseqs, char **data);


/*from random.c*/
void bootstrapSelect(int length, int *filter, int *seed);
void subsetSelect(double p, int length, int *filter, int *seed);
void sgenrand(unsigned long seed);

void counterTest(int counter, int argc)
{
  if (counter < argc)
    ;
  else
    {
      fprintf(stderr,"Error with command line: last flag called needs an argument.\n");
      exit(EXIT_FAILURE);
    }
}

void printInstructions()
{
  fprintf(stderr,"Usage: fastme -bDinopstvw\n");
  fprintf(stderr,"-b specify method for building initial tree: ");
  fprintf(stderr,"BME, GME or NJ (default).\n");
  fprintf(stderr,"-B (# replicates) if bootstrapping\n");
  fprintf(stderr,"-D (DNA-formula-type) if input consists of DNA sequences\n");
  fprintf(stderr,"-i name of input file (distance matrix by default)\n");
  fprintf(stderr,"-M filename for output matrix (none by default)\n");
  fprintf(stderr,"-n number of trees/matrices input (default = 1)\n");
  fprintf(stderr,"-o filename for tree output\n");
  fprintf(stderr,"-p if input consists of protein sequences\n");
  fprintf(stderr,"-r for (balanced OLS) SPR postprocessing\n");
  fprintf(stderr,"-s specify type of tree swapping (NNIs): ");
  fprintf(stderr,"(b)alanced, (O)LS, or (n)one. (Default is balanced.)\n");
  fprintf(stderr,"-S (scoring matrix file) if using SCOREDIST method\n");
  fprintf(stderr,"-t (optional) filename of starting tree topology\n");
  fprintf(stderr,"-v for verbose output\n");
  fprintf(stderr,"-w (b)alanced or (O)LS branch lengths (if not doing NNIs on input topology) \n");
  fprintf(stderr,"-z seed for randomization (if doing bootstrapping)\n");
  fprintf(stderr,"-help to get this message\n");
  fprintf(stderr,"(Hit return key to continue.)\n");
  fflush(stderr);
  getchar();
}



void chooseSettings(int argc, char **argv, int *btype, int *itype, int *mtype,
		    int *ntype, int *wtype, int *numDataSets, int *numReps,
		    int *seed, char **filenames, boolean *BalSPR)
{
  int counter = 1;
  sprintf(filenames[0],"input.d");
  sprintf(filenames[1],"output.tre");
  sprintf(filenames[2],"input.tre");
  sprintf(filenames[3],"output.d");
  while (counter < argc)
    {
      switch(argv[counter][1])
	{	
	case 'B':
	  counter++;
	  counterTest(counter,argc);
	  *numReps = atoi(argv[counter]);
	  if (*numReps < 1)
	    {
	      fprintf(stderr,"Error inputting number of bootstrap replicates.\n");
	      printInstructions();
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 'b':
	  counter++;
	  counterTest(counter,argc);
	  switch(argv[counter][0])
	    {
	    case 'b':
	    case 'B':
	      *btype = BAL;
	      break;
	    case 'g':
	    case 'G':
	    case 'o':
	    case 'O':
	      *btype = OLS;
	      break;
	    case 'N':
	    case 'n':
	      *btype = NJ;
	      break;
	    default:
	      fprintf(stderr,"Unknown argument to -b option: please");
	      fprintf(stderr," use BME, GME/OLS, or NJ\n");
	      printInstructions();
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 'i':
	  counter++;
	  counterTest(counter,argc);
	  if (NULL != argv[counter])
	    strcpy(filenames[0],argv[counter]);
	  else
	    {
	      fprintf(stderr,"Error: -i flag requires argument.\n");
	      printInstructions();
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 'M':
	  counter++;
	  if (NULL != argv[counter])
	    strcpy(filenames[3],argv[counter]);
	  else
	    {
	      fprintf(stderr,"Error: -M flag requires argument.\n");
	      printInstructions();
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 'n':
	  counter++;
	  *numDataSets = atoi(argv[counter]);
	  counter++;
	  break;
	case 'o':
	  counter++;
	  counterTest(counter,argc);
	  if (NULL != argv[counter])	    
	    strcpy(filenames[1],argv[counter]);
	  else
	    {
	      fprintf(stderr,"Error: -o flag requires argument.\n");
	      printInstructions();
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 't':
	  counter++;
	  counterTest(counter,argc);
	  if (NULL != argv[counter])
	    strcpy(filenames[2],argv[counter]);
	  else
	    {
	      fprintf(stderr,"Error: -t flag requires argument.\n");
	      printInstructions();
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  *btype = USER;
	  break;
	case 'w':
	  counter++;
	  counterTest(counter,argc);
	  switch(argv[counter][0])
	    {
	    case 'b':
	    case 'B':
	      *wtype = BAL;
	      break;
	    case 'g':
	    case 'G':
	    case 'o':
	    case 'O':
	      *wtype = OLS;
	      break;
	    default:
	      fprintf(stderr,"Unknown argument to -w option: please");
	      fprintf(stderr," use (b)alanced or (O)LS\n");
	      printInstructions();
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 's':
	  counter++;
	  counterTest(counter,argc);
	  switch(argv[counter][0])
	    {
	    case 'b':
	    case 'B':
	      *ntype = BAL;
	      break;
	    case 'g':
	    case 'G':
	    case 'o':
	    case 'O':
	      *ntype = OLS;
	      break;
	    case 'n':
	    case 'N':
	      *ntype = NONE;
	      break;
	    default:
	      fprintf(stderr,"Unknown argument to -s option: please");
	      fprintf(stderr," use BLS, OLS, or none\n");
	      printInstructions();
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 'D':
	  *itype = DNA;
	  counter++;
	  counterTest(counter,argc);
	  switch(argv[counter][0]) {
	  case 'K':
	    *mtype = KIMURA;
	    break;
	  case 'F':
	    *mtype = FELSENSTEIN;
	    break;
	  case 'T':
	    *mtype = TAMURA;
	    break;
	  case 'X':
	    *mtype = TRANSVERSIONONLY;
	    break;
	  case 'L':
	    *mtype = LOGDET;
	    break;
	  default:
	    fprintf(stderr,"Error inputing data type with '-D' flag.\n");
	    fprintf(stderr,"-D flag requires 'F', 'K', 'L', 'T, or 'X' as argument according to which model is desired,\n");
	    fprintf(stderr,"F84, K2P, LogDet, Tamura/Nei or XX (transversion only).\n");
	    printInstructions();
	    exit(EXIT_FAILURE);
	  }
	  counter++;
	  break;
	case 'S':
	  counter++;
	  counterTest(counter,argc);
	  *itype = SCOREDIST;
	  *mtype = SCOREDIST;
	  sprintf(filenames[4],argv[counter]);
	  counter++;
	  break;
	case 'p':
	  counter++;
	  *itype = PROTEIN;
	  *mtype = PROTEIN;
	  break;
	case 'r':
	  *BalSPR = TRUE;
	  counter++;
	  break;
	case 'v':
	  verbose = TRUE;
	  counter++;
	  break;
	case 'z':
	  counter++;
	  counterTest(counter,argc);
	  *seed = atoi(argv[counter]);
	  counter++;
	  break;
	case 'h':
	default:
	  printInstructions();	 
	  exit(0);
	}
    }
}

double **initDoubleMatrix(int d)
{
  int i,j;
  double **A;
  A = (double **) malloc(d*sizeof(double *));
  for(i=0;i<d;i++)
    {
      A[i] = (double *) malloc(d*sizeof(double));
      for(j=0;j<d;j++)
	A[i][j] = 0.0;
    }
  return(A);
}


main(int argc, char **argv)
{
  int i;
  int setCounter = 0;
  FILE *ifile1, *ifile2, *ofile1, *ofile2, *sfile;
  double **D, **A, **scoreMatrix;
  boolean BalSPR=FALSE;
  set *species, *slooper;
  node *addNode;
  char **filenames;
  char **data;
  tree *T;
  int seed;
  int repCounter, numSpecies, scoreMatrixSize;
  int seqLength = -1;
  int numSets = 1;
  int count = 0;
  int nniCount = 0;
  int sprCount = 0;
  int *filter;
  char *scoreMatrixAlphabet;
  int numReps, ntype, btype, mtype, wtype, itype;
  itype = MATRIX;
  ntype = BAL;
  wtype = OLS;
  btype = NJ;
  mtype = NONE;
  verbose = FALSE;
  filenames = (char **) malloc (5*sizeof(char *));
  T = NULL;
  scoreMatrix = NULL;
  seed = clock();
  numReps = 1;
  for(i=0;i<5;i++)
    filenames[i] = (char *) malloc(MAX_FILE_NAME_LENGTH*sizeof(char));
  chooseSettings(argc,argv,&btype,&itype,&mtype, &ntype,&wtype,&numSets,&numReps,&seed,filenames,&BalSPR);
  ifile1 = fopen(filenames[0],"r");
  if (!ifile1)
    {
      fprintf(stderr,"Error opening input file %s\n",filenames[0]);
      printInstructions();
      exit(EXIT_FAILURE);
    }
  ofile1 = fopen(filenames[1],"w");
  if (!ofile1)
    {
      fprintf(stderr,"Error opening output file %s\n",filenames[1]);
      printInstructions();
      exit(EXIT_FAILURE);
    }
  ofile2 = fopen(filenames[3],"w");
  if (!ofile2)
    {
      fprintf(stderr,"Error opening output file %s\n",filenames[3]);
      printInstructions();
      exit(EXIT_FAILURE);
    }
  if (USER == btype)
    ifile2 = fopen(filenames[2],"r");
  if (SCOREDIST == mtype)
    {
      sfile = fopen(filenames[4],"r");
      if (!(sfile))
	{
	  fprintf(stderr,"Error opening protein scoring matrix %s.\n",filenames[4]);
	  exit(EXIT_FAILURE);
	}
      scoreMatrixAlphabet = (char *)calloc(PROTEIN_ALPHABET_SIZE,sizeof(char));
      scoreMatrix = loadScoreMatrix(&scoreMatrixSize,sfile,scoreMatrixAlphabet);
    }
  sgenrand(seed);
  while ( setCounter < numSets ) {
    repCounter=0;
    species = (set *) malloc(sizeof(set));
    species->firstNode = NULL;
    species->secondNode = NULL;
    if (0 == (setCounter +1) % 100) 
      printf("*");
    else
      printf(".");
    fflush(stdout);
    
    if (MATRIX != itype)
	data = loadSequences(ifile1,&numSpecies,species,&seqLength);
    else 
      D = loadMatrix(ifile1,&numSpecies,species);
    while (repCounter < numReps)
      {
	if (numReps > 1)
	  {
	    printf(",");
	    fflush(stdout);
	  }
	if (MATRIX != itype)
	  { /*sequence input*/
	    filter = (int *)malloc(seqLength*sizeof(int));
	    if (numReps > 1)
	      {
		filter = initZeroArray(seqLength);
		bootstrapSelect(seqLength,filter,&seed);
	      }
	    else
	      filter = initOneArray(seqLength);
	    gapCheckFilter(filter,itype,seqLength,numSpecies,data);
	    D = makeDistMatrix(data, numSpecies, seqLength, mtype, filter,1.0,scoreMatrix,scoreMatrixSize,scoreMatrixAlphabet);
	    printMatrix(D,numSpecies,species,ofile2);
	    free(filter);
	  }
	
	/*using the matrix to build a tree*/
	A = initDoubleMatrix(2*numSpecies-2);  
	switch(btype)
	  {
	  case USER:
	    if (!ifile2)
	      {
		fprintf(stderr,"Error opening input file %s\n",filenames[2]);
		printInstructions();
		exit(EXIT_FAILURE);
	      }
	    T = loadNewickTree(ifile2,numSpecies);
	    T = detrifurcate(T);      
	    compareSets(T,species,ofile1);
	    partitionSizes(T);
	    break;
	  case OLS:
	    for(slooper = species; NULL != slooper; slooper = slooper->secondNode)
	      {
		addNode = copyNode(slooper->firstNode);
		T = GMEaddSpecies(T,addNode,D,A);
	      }
	    break;
	  case BAL:
	    for(slooper = species; NULL != slooper; slooper = slooper->secondNode)
	      {
		addNode = copyNode(slooper->firstNode);
		T = BMEaddSpecies(T,addNode,D,A);
	      }
	    break;
	  case NJ:
	    T = nj(D,species,numSpecies);
	    break;
	  }
	switch(ntype)
	  {      
	  case OLS:
	    if (OLS != btype)
	      assignAllSizeFields(T);
	    makeOLSAveragesTable(T,D,A);
	    NNI(T,A,&nniCount);
	    assignOLSWeights(T,A);
	    break;
	  case BAL:	
	    if (BAL != btype)
	      makeBMEAveragesTable(T,D,A);
	    bNNI(T,A,&nniCount);
	    assignBMEWeights(T,A);
	    break;
	  case NONE:
	    switch(wtype)
	      {
	      case OLS:
		if (OLS != btype)
		  assignAllSizeFields(T);
		makeOLSAveragesTable(T,D,A);
		assignOLSWeights(T,A);
		break;
	      case BAL:
		if (BAL != btype)
		  makeBMEAveragesTable(T,D,A);
		assignBMEWeights(T,A);
		break;
	      default:
		fprintf(stderr,"Error in program: variable 'btype' has illegal ");
		fprintf(stderr,"value %d.\n",btype);
		exit(EXIT_FAILURE);
	      }
	    break;
	  default:
	    fprintf(stderr,"Error in program: variable 'ntype' has illegal ");
	    fprintf(stderr,"value %d.\n",ntype);
	    exit(EXIT_FAILURE);
	  }
	if (BalSPR)
	  {
	    if ((OLS == ntype) || ((NONE==ntype)&&(BAL!=btype)))
	      makeBMEAveragesTable(T,D,A);	    
	    SPR(T,D,A,&sprCount);
	    assignBMEWeights(T,A);
	  }
	NewickPrintTree(T,ofile1);
	freeMatrix(D,numSpecies);
	freeMatrix(A,2*numSpecies - 2);
	freeTree(T);
	T = NULL;
	setCounter++;
	if ((verbose) && ((ntype) || (BalSPR)))
	  printf("Performed %d NNIs and %d SPRs on data set %d\n",nniCount,sprCount,setCounter);
	nniCount = sprCount = 0;
	repCounter++;
      } /*end bootstrap loop*/
    freeSet(species);
    if (MATRIX != itype)
      freeCharMatrix(data,numSpecies);
  }
  printf("\n");
  fclose(ifile1);
  if (USER == btype)
    fclose(ifile2);
  fclose(ofile1);
  fclose(ofile2);
  freeCharMatrix(filenames,3);
}
