#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fastme.h"
#include "graph.h"
#include "distance.h"
#include <string.h>

boolean leaf(node *v);

int *initPerm(int n);

double **initDoubleMatrix(int d);
void freeDoubleMatrix(double **X, int d);

edge *depthFirstTraverse(tree *T, edge *e);

void compareSets(tree *T, set *S, FILE *ofile)
{
  edge *e;
  node *v,*w;
  set *X;
  e = depthFirstTraverse(T,NULL);
  while (NULL != e)
    {
      v = e->head;
      for(X = S; NULL != X; X = X->secondNode)
	{
	  w = X->firstNode;
	  if (0 == strcmp(v->label,w->label))
	    {
	      v->index2 = w->index2;
	    w->index2 = -1;
	    break;
	    }
	}
      e = depthFirstTraverse(T,e);
    }
  v = T->root;
  for(X = S; NULL != X; X = X->secondNode)
    {
      w = X->firstNode;
      if (0 == strcmp(v->label,w->label))
	{
	  v->index2 = w->index2;
	  w->index2 = -1;
	  break;
	}
    }
  if (-1 == v->index2)
    {
      fprintf(stderr,"Error leaf %s in tree not in distance matrix.\n",v->label);
      exit(EXIT_FAILURE);
}
  e = depthFirstTraverse(T,NULL);
  while (NULL != e)
    {
      v = e->head;
      if ((leaf(v)) && (-1 == v->index2))
	{
	  fprintf(stderr,"Error leaf %s in tree not in distance matrix.\n",v->label);
	  exit(EXIT_FAILURE);
	}
      e = depthFirstTraverse(T,e);
      }
  for(X = S; NULL != X; X = X->secondNode)
    if (X->firstNode->index2 > -1)
      {
	fprintf(ofile,"(v1:0.0)v2;");
	fclose(ofile);
	fprintf(stderr,"Error node %s in matrix but not a leaf in tree.\n",X->firstNode->label);
	exit(EXIT_FAILURE);
      }
}

void freeCharMatrix(char **D, int size)
{
  int i;
  for(i=0;i<size;i++)
    free(D[i]);
  free(D);
}

void freeMatrix(double **D, int size)
{
  int i;
  for(i=0;i<size;i++)
    free(D[i]);
  free(D);
}

double **loadMatrix(FILE *ifile, int *size, set *S)
{
  char nextString[MAX_EVENT_NAME];
  node *v;
  double **table;
  double val;
  int i,j;
  if (!(fscanf(ifile,"%s",nextString)))
    {
      fprintf(stderr,"Error loading input matrix.\n");
      exit(EXIT_FAILURE);
    }
  *size = atoi(nextString);
  if ((*size < 0) || (*size > MAXSIZE))
    {
      printf("Problem inputting size.\n");
      exit(EXIT_FAILURE);
    }
  table = (double **) calloc(*size,sizeof(double *));
  for(i=0;i<*size;i++)
    {
      j = 0;
      table[i] = (double *) calloc(*size,sizeof(double));
      if (!(fscanf(ifile,"%s",nextString)))
	{
	  fprintf(stderr,"Error loading label %d.\n",i);
	  exit(EXIT_FAILURE);
	}
      v = makeNewNode(nextString,-1);
      v->index2 = i;
      S = addToSet(v,S);
      while (j < *size)
	{
	  if (!(fscanf(ifile,"%s",nextString)))
	    {
	      fprintf(stderr,"Error loading (%d,%d)-entry.\n",i,j);
	      exit(EXIT_FAILURE);
	    }
	  val = atof(nextString);
	  if ((i != j) && (0 > val))
	    {
	      fprintf(stderr,"Error: program is expecting distance matrix.");
	      fprintf(stderr,"\nInput of %s off diagonal is inappropriate.\n",nextString);
	      fprintf(stderr,"Please use -D, -p, or -S if using sequence data.\n");
	      exit(EXIT_FAILURE);
	    }
	  table[i][j++] = val;	  
	}	  
    }
  return(table);
}

void partitionSizes(tree *T)
{
  edge *e;
  e = depthFirstTraverse(T,NULL);
  while (NULL != e)
    {
      if (leaf(e->head))
	e->bottomsize = 1;
      else
	e->bottomsize = e->head->leftEdge->bottomsize 
	  + e->head->rightEdge->bottomsize;
      e->topsize = (T->size + 2)/2 - e->bottomsize;
      e = depthFirstTraverse(T,e);
    }
}

char **loadSequences(FILE *ifile, int *numSeqs, set *taxa, int *inputSize)
{
  int i;
  node *v;
  char *nextString;
  char **data;
  nextString = (char *)malloc(MAX_LABEL_LENGTH*sizeof(char));  
  if (!(fscanf(ifile,"%s",nextString)))
    {
      fprintf(stderr,"Error loading input file.\n");
      exit(EXIT_FAILURE);
    }
  *numSeqs = atoi(nextString);
  fscanf(ifile,"%s",nextString);
  *inputSize = atoi(nextString);
  if (*inputSize < 0)
    {
      fprintf(stderr,"Error loading sequence length from input file.\n");
      exit(EXIT_FAILURE);
    }
  if (verbose)
    printf("numSeqs is %d\n",*numSeqs);
  data = (char **)calloc(*numSeqs,sizeof(char *));
  if (verbose)
    printf("inputSize is %d\n",*inputSize);
  for(i=0;i<*numSeqs;i++)
    {
      if (!(fscanf(ifile,"%s",nextString)))
	{
	  fprintf(stderr,"Error reading label of sequence %d.\n",i+1);
	  exit(EXIT_FAILURE);
	}
      v = makeNewNode(nextString,-1);
      v->index2 = i;
      taxa = addToSet(v,taxa);      
      data[i] = (char *)calloc(*inputSize+MAX_LABEL_LENGTH,sizeof(char));
      fscanf(ifile,"%s",data[i]);
    }
  free(nextString);
  return(data);
}

double **loadScoreMatrix(int *d, FILE *ifile, char *alphabet)
{
  char nextString[MAX_EVENT_NAME];
  double **table;
  int i,j;
  if (!(fscanf(ifile,"%s",nextString)))
    {
      fprintf(stderr,"Error loading size of score matrix.\n");
      exit(EXIT_FAILURE);
    }
  *d = atoi(nextString);
  if ((*d < 1) || (*d > PROTEIN_ALPHABET_SIZE))
   {
      fprintf(stderr,"Error loading size of score matrix.\n");
      exit(EXIT_FAILURE);
    }
  table = initDoubleMatrix(*d);
  /* *d is actual size of input matrix, which may be less than PROTEIN_ALPHABET_SIZE, 
     since the ALPHABET contains special characters*/
  for(i=0;i<*d;i++)
    {
      j = 0;
      if (!(fscanf(ifile,"%s",nextString)))
	{
	  fprintf(stderr,"Error loading label %d.\n",i);
	  exit(EXIT_FAILURE);
	}
      alphabet[i] = nextString[0];
      while (j < *d)
	{
	  if (!(fscanf(ifile,"%s",nextString)))
	    {
	      fprintf(stderr,"Error loading (%d,%d)-entry.\n",i,j);
	      exit(EXIT_FAILURE);
	    }
	  table[i][j++] = atof(nextString);
	} /*j loop*/	  
    }  /*i loop*/
  if (*d < PROTEIN_ALPHABET_SIZE)
    alphabet[*d]='\0';
  return(table);
}
